# pyright: reportMissingTypeStubs=none

import jproperties
import pytest

from hatch_gradle_version.common.gradle import GradleVersion


def test_from_properties_valid():
    p = jproperties.Properties()
    p["key"] = "0.11.1-7"

    gradle_version = GradleVersion.from_properties(p, "key", None)

    assert gradle_version == GradleVersion(
        raw_version="0.11.1-7",
        version="0.11.1",
        rc=7,
        build=None,
        extra_versions={"key": "0.11.1-7"},
    )


def test_from_properties_missing_key():
    p = jproperties.Properties()
    p["key"] = "0.11.1-7"

    with pytest.raises(KeyError, match="MISSING_KEY"):
        GradleVersion.from_properties(p, "MISSING_KEY", None)



import os
from contextlib import contextmanager
from typing import AnyStr


# https://stackoverflow.com/a/24176022
@contextmanager
def cd(newdir: os.PathLike[AnyStr]):
    """Context manager which temporarily changes the script's working directory."""
    prevdir = os.getcwd()
    os.chdir(os.path.expanduser(newdir))
    try:
        yield
    finally:
        os.chdir(prevdir)



import textwrap
from pathlib import Path

import pytest
from pytest import MonkeyPatch

from hatch_gradle_version.common.gradle import GradleVersion

from .gradle_properties import GradlePropertiesVersionSource


def test_gradle_version(tmp_path: Path, monkeypatch: MonkeyPatch):
    # arrange
    monkeypatch.setenv("HATCH_GRADLE_DIR", "gradle_dir")

    gradle = tmp_path / "gradle_dir" / "gradle.properties"
    gradle.parent.mkdir()
    gradle.write_text("modVersion=0.11.1-7")

    (tmp_path / "__version__.py").write_text('PY_VERSION = "1.0"')

    hook = GradlePropertiesVersionSource.from_config(
        tmp_path.as_posix(),
        {
            "source": "json",
            "py-path": "__version__.py",
            "gradle-path": "gradle.properties",
            "key": "modVersion",
        },
    )

    gradle_version = hook.get_gradle_version()

    assert gradle_version == GradleVersion(
        raw_version="0.11.1-7",
        version="0.11.1",
        rc=7,
        build=None,
        extra_versions={"modVersion": "0.11.1-7"},
    )

    # TODO: this should really be a separate test case

    hook.get_version_data()

    generated_version_file = (tmp_path / "__gradle_version__.py").read_text("utf-8")
    assert generated_version_file == textwrap.dedent(
        """\
        # This file is auto-generated by hatch-gradle-version. Do not edit.

        GRADLE_VERSION = "0.11.1-7"
        FULL_VERSION = "0.11.1.1.0rc7"

        MOD_VERSION = "0.11.1-7"
        """
    )


def test_gloopy(tmp_path: Path, monkeypatch: MonkeyPatch):
    # arrange
    monkeypatch.setenv("HATCH_GRADLE_DIR", "gradle_dir")

    gradle = tmp_path / "gradle_dir" / "gradle.properties"
    gradle.parent.mkdir()
    gradle.write_text("modVersion=1.19.2-0.1.0")

    hook = GradlePropertiesVersionSource.from_config(
        tmp_path.as_posix(),
        {
            "source": "json",
            "py-path": "",
            "gradle-path": "gradle.properties",
            "key": "modVersion",
            "gradle-version-regex": r".+?-(\d+\.\d+\.\d+)",
        },
    )

    gradle_version = hook.get_gradle_version()

    assert gradle_version == GradleVersion(
        raw_version="0.1.0",
        version="0.1.0",
        rc=None,
        build=None,
        extra_versions={"modVersion": "1.19.2-0.1.0"},
    )


@pytest.mark.parametrize(
    ["pattern", "replacement", "raw_version", "expected"],
    [
        (r"^(.+?)-(.+)", r"\2.\1", "1.19.2-0.1.0", "0.1.0.1.19.2"),
        (r"^(.+?)\+(.+)", r"\1.\2", "0.1.0+1.19.2", "0.1.0.1.19.2"),
        (r"(.+)", r"\1.{minecraftVersion}", "0.1.0", "0.1.0.1.19.2"),
    ],
)
def test_rewrite(
    tmp_path: Path,
    monkeypatch: MonkeyPatch,
    pattern: str,
    replacement: str,
    raw_version: str,
    expected: str,
):
    # arrange
    monkeypatch.setenv("HATCH_GRADLE_DIR", "gradle_dir")

    gradle = tmp_path / "gradle_dir" / "gradle.properties"
    gradle.parent.mkdir()
    gradle.write_text(
        textwrap.dedent(
            f"""\
            modVersion={raw_version}
            minecraftVersion=1.19.2
            """
        )
    )

    hook = GradlePropertiesVersionSource.from_config(
        tmp_path.as_posix(),
        {
            "source": "json",
            "py-path": "",
            "gradle-path": "gradle.properties",
            "key": "modVersion",
            "gradle-version-regex": {
                "pattern": pattern,
                "replacement": replacement,
            },
        },
    )

    gradle_version = hook.get_gradle_version()

    assert gradle_version == GradleVersion(
        raw_version=expected,
        version=expected,
        rc=None,
        build=None,
        extra_versions={
            "modVersion": raw_version,
            "minecraftVersion": "1.19.2",
        },
    )



from pathlib import Path

from hatch_gradle_version.common.gradle import GradleVersion, load_properties
from hatch_gradle_version.common.model import GradlePath

from .base import BaseVersionSource


class GradlePropertiesVersionSource(BaseVersionSource):
    PLUGIN_NAME = "gradle-properties"

    gradle_path: GradlePath = Path("gradle.properties")
    key: str = "modVersion"

    def get_gradle_version(self) -> GradleVersion:
        p = load_properties(self.gradle_path)
        return GradleVersion.from_properties(p, self.key, self.fmt_raw_gradle_version)



from pathlib import Path
from textwrap import dedent

import black
from black.mode import Mode


def write_code(path: str | Path, *lines: str):
    raw_code = "\n".join(dedent(line) for line in lines)
    formatted_code = black.format_str(raw_code, mode=Mode())
    Path(path).write_text(formatted_code, "utf-8")



from __future__ import annotations

import re
import warnings
from pathlib import Path
from typing import Callable

import jproperties  # pyright: ignore[reportMissingTypeStubs]
from packaging.version import Version

from .model import DefaultModel, KebabModel

GRADLE_VERSION_RE = re.compile(
    r"""
    v?
    (?P<version>
        \d+
        (?:\.\d+)*
    )
    (?:
        -(?P<rc>\d+)
    )?
    (?:
        \+(?P<build>.+)
    )?
    """,
    re.VERBOSE,
)


# this uses Model so I can be lazy with constructing it from a regex
class GradleVersion(DefaultModel, arbitrary_types_allowed=True):
    raw_version: str
    version: str
    rc: int | None
    build: str | None
    extra_versions: dict[str, str]

    @classmethod
    def from_properties(
        cls,
        p: jproperties.Properties,
        key: str,
        fmt: Callable[[str, dict[str, str]], str] | None,
    ):
        if key not in p:
            raise KeyError(f"Key not found in gradle.properties: {key}")
        return cls.from_raw(
            raw_version=str(p[key].data),
            extra_versions={key: value.data for key, value in p.items()},
            fmt=fmt,
        )

    @classmethod
    def from_raw(
        cls,
        raw_version: str,
        extra_versions: dict[str, str],
        fmt: Callable[[str, dict[str, str]], str] | None,
    ):
        if fmt:
            raw_version = fmt(raw_version, extra_versions)

        match = GRADLE_VERSION_RE.match(raw_version)
        if match is None:
            raise ValueError(f"Failed to parse version: {raw_version}")

        data = match.groupdict() | {
            "raw_version": raw_version,
            "extra_versions": extra_versions,
        }
        return cls.model_validate(data)

    def full_version(
        self,
        py_version: str | Version,
        *,
        next_rc: bool = False,
    ) -> str:
        # ensure py_version is a valid version by itself
        if isinstance(py_version, str):
            py_version = Version(py_version)

        if py_version.pre:
            raise ValueError("a/b/c/pre/rc is reserved for Gradle prereleases")

        # split py_version at the point where we need to insert the gradle prerelease
        py_base = py_version.base_version
        py_rest = str(py_version).removeprefix(py_base)

        # construct the full version
        # eg. 1.2.3 . 4.5 rc6 .dev7
        rc = self.rc_segment(next_rc)
        full_version = self.version + "." + py_base + rc + py_rest

        # round-trip through Version to normalize it
        return str(Version(full_version))

    def rc_segment(self, next_rc: bool):
        if self.rc is None:
            return ""
        return f"rc{self.next_rc if next_rc else self.rc}"

    @property
    def next_rc(self):
        if self.rc is None:
            raise ValueError("Tried to call next_rc on a non-rc version")
        return self.rc + 1

    def __str__(self) -> str:
        return self.raw_version


def load_properties(path: Path):
    p = jproperties.Properties()
    with open(path, "rb") as f:
        p.load(f, "utf-8")
    return p


class GradleDependency(KebabModel):
    package: str
    op: str
    key: str
    py_version: str
    rc_upper_bound: bool = False
    """If True and gradle_version has a pre-release suffix (eg. `0.1.0-1`), add a
    corresponding exclusive upper bound for the next RC version (eg. `<0.1.0.1.0rc2`).
    
    If True, the operators `>=` and `~=` effectively become `==`. If False, pip may
    install a later prerelease or a released version. There's not really a "best"
    option, which is why this flag exists.

    The default is `False`. This *should* work in more cases than not.
    """

    def version_specifier(self, gradle_version: GradleVersion):
        full_version = gradle_version.full_version(self.py_version)
        lower_bound = self.op + full_version

        if gradle_version.rc is None or not self.rc_upper_bound:
            return self.package + lower_bound

        if "<" not in self.op:
            warnings.warn(
                f"Dependency on package `{self.package}` will ONLY accept `{full_version}` (`{gradle_version}` is a prerelease and `rc-upper-bound` is enabled)."
            )

        upper_bound = "<" + gradle_version.full_version(self.py_version, next_rc=True)
        return f"{self.package}{lower_bound},{upper_bound}"



from typing import Any

from hatchling.version.scheme.standard import StandardScheme

from ..common.gradle import GradleVersion
from .version_source.base import VersionData


class GradleVersionScheme(StandardScheme):
    PLUGIN_NAME = "gradle"

    def update(
        self,
        desired_version: str,
        original_version: str,
        version_data: dict[str, Any] | VersionData,
    ) -> str:
        assert isinstance(version_data["py_version"], str)
        assert isinstance(version_data["full_gradle_version"], GradleVersion)

        # just update the Python component of the version
        new_py_version = super().update(
            desired_version=desired_version,
            original_version=version_data["py_version"],
            version_data=version_data | {"version": version_data["py_version"]},
        )

        # update version_data so version_source can write the updated py version
        version_data["py_version"] = new_py_version
        return version_data["full_gradle_version"].full_version(new_py_version)



from functools import wraps
from typing import Callable, Iterator, ParamSpec, TypeVar

_P = ParamSpec("_P")
_Item = TypeVar("_Item")


def listify(func: Callable[_P, Iterator[_Item]]) -> Callable[_P, list[_Item]]:
    """Decorator for making a generator return a list instead."""

    @wraps(func)
    def wrapper(*args: _P.args, **kwargs: _P.kwargs):
        return list(func(*args, **kwargs))

    return wrapper



import os
from pathlib import Path
from typing import Annotated, Any, ClassVar

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    ValidationError,
    ValidationInfo,
    field_validator,
    model_validator,
)

DEFAULT_CONFIG = ConfigDict(extra="forbid")


def to_kebab(string: str) -> str:
    return string.replace("_", "-")


class DefaultModel(BaseModel):
    model_config = DEFAULT_CONFIG

    # workaround for https://github.com/pypa/hatch/issues/959
    @classmethod
    def model_validate(
        cls,
        obj: Any,
        *,
        strict: bool | None = None,
        from_attributes: bool | None = None,
        context: dict[str, Any] | None = None,
    ):
        try:
            return super().model_validate(
                obj,
                strict=strict,
                from_attributes=from_attributes,
                context=context,
            )
        except ValidationError as e:
            # wrap ValidationError so Hatchling doesn't try to construct it
            raise RuntimeError(e)

    # workaround for https://github.com/pypa/hatch/issues/959
    @classmethod
    def model_validate_json(
        cls,
        json_data: str | bytes | bytearray,
        *,
        strict: bool | None = None,
        context: dict[str, Any] | None = None,
    ):
        try:
            return super().model_validate_json(
                json_data,
                strict=strict,
                context=context,
            )
        except ValidationError as e:
            raise RuntimeError(e)


class KebabModel(DefaultModel, alias_generator=to_kebab):
    pass


ProjectPath = Annotated[Path, "ProjectPath"]

GradlePath = Annotated[Path, "GradlePath"]


class HookModel(KebabModel, validate_default=True):
    PLUGIN_NAME: ClassVar[str]

    root__: str | Path = Field(alias="root", kw_only=False)
    config__: dict[str, Any] = Field(alias="config", kw_only=False)

    def __init__(self, root: str | Path, config: dict[str, Any]):
        self.__pydantic_validator__.validate_python(
            {"root": root, "config": config},
            self_instance=self,
        )

    @classmethod
    def from_config(cls, root: str | Path, config: dict[str, Any]):
        return cls(root, config)

    @property
    def root(self):
        return self.root__

    @property
    def config(self):
        return self.config__

    @model_validator(mode="before")
    def _merge_with_config(cls, value: Any):
        match value:
            case {"config": dict() as config}:
                return config | value
            case _:
                return value

    @field_validator("*", mode="after")
    @classmethod
    def _resolve_paths(cls, value: Any, info: ValidationInfo):
        root = info.data.get("root__")
        if not root:
            return value
        root = Path(root)

        assert info.field_name
        field_info = cls.model_fields[info.field_name]

        for annotation in field_info.metadata:
            match annotation:
                case "ProjectPath":
                    return root / value
                case "GradlePath":
                    return root / os.getenv("HATCH_GRADLE_DIR", "") / value
                case _:
                    pass

        return value



import copy
import textwrap
from pathlib import Path

import pytest
from pytest import MonkeyPatch

from .gradle_properties import GradlePropertiesMetadataHook


@pytest.mark.parametrize(
    # fmt: off
    "package, op,key,py_version,gradle_version,rc_upper_bound,full_version",
    [
        ("P", "~=", "KEY", "4",       "1.2.3",   False, "P~=1.2.3.4"),
        ("P", "~=", "KEY", "4.5",     "1.2.3",   False, "P~=1.2.3.4.5"),
        ("P", ">=", "KEY", "4.5",     "1.2.3",   False, "P>=1.2.3.4.5"),
        ("P", "~=", "KEY", "4.5",     "1.2.3-6", False, "P~=1.2.3.4.5rc6"),
        ("P", "~=", "KEY", "4.5",     "1.2.3-6", True,  "P~=1.2.3.4.5rc6,<1.2.3.4.5rc7"),
        ("P", "~=", "KEY", "4.5dev6", "1.2.3",   False, "P~=1.2.3.4.5.dev6"),
        ("P", "~=", "KEY", "4.5dev7", "1.2.3-6", False, "P~=1.2.3.4.5rc6.dev7"),
        ("P", "~=", "KEY", "4.5dev8", "1.2.3-6", True,  "P~=1.2.3.4.5rc6.dev8,<1.2.3.4.5rc7.dev8"),
    ],
    # fmt: on
)
def test_gradle_properties_deps(
    tmp_path: Path,
    package: str,
    op: str,
    key: str,
    py_version: str,
    gradle_version: str,
    rc_upper_bound: bool,
    full_version: str,
    monkeypatch: MonkeyPatch,
):
    # arrange
    monkeypatch.setenv("HATCH_GRADLE_DIR", "gradle_dir")

    gradle_properties = tmp_path / "gradle_dir" / "gradle.properties"
    gradle_properties.parent.mkdir()
    gradle_properties.write_text(f"{key}={gradle_version}")

    hook = GradlePropertiesMetadataHook(
        tmp_path.as_posix(),
        {
            "dependencies": [
                {
                    "package": package,
                    "op": op,
                    "key": key,
                    "py-version": py_version,
                    "rc-upper-bound": rc_upper_bound,
                }
            ],
        },
    )

    orig_metadata = {
        "dynamic": [
            "dependencies",
            "optional-dependencies",
        ],
    }

    # act
    metadata = copy.deepcopy(orig_metadata)
    hook.update(metadata)

    # assert
    assert metadata == orig_metadata | {
        "dependencies": [full_version],
        "optional-dependencies": {},
    }


@pytest.mark.parametrize(
    ["dependency", "result"],
    [
        ("", ""),
        ("foo", "foo"),
        ("{foo", "{foo"),
        ("{foo}", "{foo}"),
        ("{placeholder}", "result"),
        ("{{placeholder}}", "{result}"),
        ("{placeholder}{placeholder}", "resultresult"),
        ("{placeholder} {other}", "result value"),
        ("{other} {placeholder}", "value result"),
        ("foo bar baz", "foo bar baz"),
        ("{foo} bar baz", "{foo} bar baz"),
        ("foo {bar} baz", "foo {bar} baz"),
        ("foo bar {baz}", "foo bar {baz}"),
        ("{placeholder} bar baz", "result bar baz"),
        ("foo {placeholder} baz", "foo result baz"),
        ("foo bar {placeholder}", "foo bar result"),
        (
            "package @ {root:uri}/path/to/package-0.1.0-py3-none-any.whl",
            "package @ {root:uri}/path/to/package-0.1.0-py3-none-any.whl",
        ),
    ],
)
def test_placeholders(
    tmp_path: Path,
    monkeypatch: MonkeyPatch,
    dependency: str,
    result: str,
):
    # arrange
    monkeypatch.setenv("HATCH_GRADLE_DIR", "gradle_dir")

    gradle_properties = tmp_path / "gradle_dir" / "gradle.properties"
    gradle_properties.parent.mkdir()
    gradle_properties.write_text(
        textwrap.dedent(
            """\
            placeholder=result
            other=value
            """
        )
    )

    hook = GradlePropertiesMetadataHook(
        tmp_path.as_posix(),
        {
            "dependencies": [dependency],
            "optional-dependencies": {"dev": [dependency]},
        },
    )

    metadata = {
        "dynamic": [
            "dependencies",
            "optional-dependencies",
        ],
    }

    # act
    hook.update(metadata)

    # assert
    assert metadata == {
        "dynamic": [
            "dependencies",
            "optional-dependencies",
        ],
        "dependencies": [result],
        "optional-dependencies": {"dev": [result]},
    }



import json
from typing import Annotated, Any, overload

from pydantic import Field, field_validator
from pydantic.types import JsonValue

from hatch_gradle_version.common.gradle import GradleVersion
from hatch_gradle_version.common.model import GradlePath

from .base import BaseVersionSource


class JSONVersionSource(BaseVersionSource):
    PLUGIN_NAME = "json"

    json_path: GradlePath
    key: Annotated[list[str], Field(min_length=1)]

    def get_gradle_version(self) -> GradleVersion:
        with self.json_path.open() as f:
            data: JsonValue = json.load(f)

        raw_version = data
        for key in self.key:
            if not isinstance(raw_version, dict):
                raise ValueError(f"Invalid key: {self.key} (not a dict: {raw_version})")
            raw_version = raw_version[key]

        raw_version = self.json_to_string(raw_version)
        if raw_version is None:
            raise ValueError(f"Invalid key: {self.key} (invalid value: {raw_version})")

        extra_versions = dict[str, str]()
        self.get_extra_versions(data, extra_versions, [])

        return GradleVersion.from_raw(
            raw_version,
            extra_versions,
            self.fmt_raw_gradle_version,
        )

    def get_extra_versions(
        self,
        value: JsonValue,
        result: dict[str, str],
        parents: list[str],
    ):
        match value:
            case dict():
                for key, child in value.items():
                    parents.append(key)
                    self.get_extra_versions(child, result, parents)
                    parents.pop()
            case list():
                for i, child in enumerate(value):
                    parents.append(str(i) if parents else f"_{i}")
                    self.get_extra_versions(child, result, parents)
                    parents.pop()
            case _:
                result[".".join(parents)] = self.json_to_string(value)

    @overload
    def json_to_string(self, value: str | bool | int | float | None) -> str: ...

    @overload
    def json_to_string(self, value: list[JsonValue] | dict[str, JsonValue]) -> None: ...

    @overload
    def json_to_string(self, value: JsonValue) -> str | None: ...

    def json_to_string(self, value: JsonValue):
        match value:
            case dict() | list():
                return None
            case True:
                return "true"
            case False:
                return "false"
            case None:
                return "null"
            case _:
                return str(value)

    @field_validator("key", mode="before")
    @classmethod
    def _split_key(cls, value: Any):
        match value:
            case str():
                return value.split(".")
            case _:
                return value



import textwrap
from pathlib import Path
from textwrap import dedent

from pytest import MonkeyPatch

from hatch_gradle_version.common.gradle import GradleVersion

from .json import JSONVersionSource


def test_json_version(tmp_path: Path, monkeypatch: MonkeyPatch):
    # arrange
    monkeypatch.setenv("HATCH_GRADLE_DIR", "gradle_dir")

    json = tmp_path / "gradle_dir" / "mod.json"
    json.parent.mkdir()
    json.write_text(
        dedent(
            """
            {
                "core": {
                    "name": "hexbound",
                    "version": "0.1.4+1.19.2",
                    "repository": "https://github.com/Cypher121/hexbound/"
                },
                "platforms": {
                    "modrinth": {
                        "id": "PHgo4bVw"
                    },
                    "curseforge": {
                        "id": "739791"
                    }
                },
                "foo": [
                    "bar",
                    "baz",
                    {
                        "qux": "quux"
                    }
                ]
            }
            """
        )
    )

    (tmp_path / "__version__.py").write_text('PY_VERSION = "1.0"')

    hook = JSONVersionSource.from_config(
        tmp_path.as_posix(),
        {
            "source": "json",
            "py-path": "__version__.py",
            "json-path": "mod.json",
            "key": "core.version",
        },
    )

    gradle_version = hook.get_gradle_version()

    assert gradle_version == GradleVersion(
        raw_version="0.1.4+1.19.2",
        version="0.1.4",
        rc=None,
        build="1.19.2",
        extra_versions={
            "core.name": "hexbound",
            "core.version": "0.1.4+1.19.2",
            "core.repository": "https://github.com/Cypher121/hexbound/",
            "platforms.modrinth.id": "PHgo4bVw",
            "platforms.curseforge.id": "739791",
            "foo.0": "bar",
            "foo.1": "baz",
            "foo.2.qux": "quux",
        },
    )

    hook.get_version_data()

    generated_version_file = (tmp_path / "__gradle_version__.py").read_text("utf-8")
    assert generated_version_file == textwrap.dedent(
        """\
        # This file is auto-generated by hatch-gradle-version. Do not edit.

        GRADLE_VERSION = "0.1.4+1.19.2"
        FULL_VERSION = "0.1.4.1.0"

        CORE_NAME = "hexbound"
        CORE_REPOSITORY = "https://github.com/Cypher121/hexbound/"
        CORE_VERSION = "0.1.4+1.19.2"
        FOO_0 = "bar"
        FOO_1 = "baz"
        FOO_2_QUX = "quux"
        PLATFORMS_CURSEFORGE_ID = "739791"
        PLATFORMS_MODRINTH_ID = "PHgo4bVw"
        """
    )



import tomllib
from functools import cached_property
from pathlib import Path
from typing import Any

from pydantic import TypeAdapter

from hatch_gradle_version.common.gradle import GradleDependency, GradleVersion
from hatch_gradle_version.common.model import GradlePath

from .base import BaseMetadataHook


class VersionCatalogMetadataHook(BaseMetadataHook):
    PLUGIN_NAME = "version-catalog"

    path: GradlePath = Path("gradle/libs.versions.toml")

    def get_format_value(self, key: str):
        return self.versions.get(key)

    def parse_gradle_dependency(self, dependency: GradleDependency):
        raw_version = self.versions[dependency.key]
        if not isinstance(raw_version, str):
            raise ValueError(
                f"Expected str for dependency {dependency.key}, got {raw_version}"
            )
        gradle_version = GradleVersion.from_raw(raw_version, self.str_versions, None)
        return dependency.version_specifier(gradle_version)

    @cached_property
    def versions(self) -> dict[str, str | dict[str, Any]]:
        with self.path.open("rb") as f:
            version_catalog = tomllib.load(f)

        ta = TypeAdapter(dict[str, str | dict[str, Any]])
        return ta.validate_python(version_catalog["versions"])

    @cached_property
    def str_versions(self):
        return {k: v for k, v in self.versions.items() if isinstance(v, str)}



from hatchling.plugin import hookimpl

from .plugins.metadata_hook.gradle_properties import GradlePropertiesMetadataHook
from .plugins.metadata_hook.version_catalog import VersionCatalogMetadataHook
from .plugins.version_scheme import GradleVersionScheme
from .plugins.version_source.gradle_properties import GradlePropertiesVersionSource
from .plugins.version_source.json import JSONVersionSource


@hookimpl
def hatch_register_version_source():
    return [GradlePropertiesVersionSource, JSONVersionSource]


@hookimpl
def hatch_register_version_scheme():
    return GradleVersionScheme


@hookimpl
def hatch_register_metadata_hook():
    return [GradlePropertiesMetadataHook, VersionCatalogMetadataHook]



from functools import cached_property
from pathlib import Path
from typing import Any

from hatch_gradle_version.common.gradle import (
    GradleDependency,
    GradleVersion,
    load_properties,
)
from hatch_gradle_version.common.model import GradlePath

from .base import BaseMetadataHook


class GradlePropertiesMetadataHook(BaseMetadataHook):
    PLUGIN_NAME = "gradle-properties"

    path: GradlePath = Path("gradle.properties")

    def get_format_value(self, key: str) -> Any | None:
        if prop := self.properties.get(key):
            return prop.data

    def parse_gradle_dependency(self, dependency: GradleDependency):
        gradle_version = GradleVersion.from_properties(
            p=self.properties,
            key=dependency.key,
            fmt=None,
        )
        return dependency.version_specifier(gradle_version)

    @cached_property
    def properties(self):
        return load_properties(self.path)



import re
from abc import ABC, abstractmethod
from typing import Any, Iterator

from hatchling.metadata.plugin.interface import MetadataHookInterface
from pydantic import Field

from hatch_gradle_version.common.decorators import listify
from hatch_gradle_version.common.gradle import GradleDependency
from hatch_gradle_version.common.model import GradlePath, HookModel

Dependencies = list[str | GradleDependency]

PLACEHOLDER_REGEX = re.compile(r"\{([a-zA-Z0-9_.\-]+)\}")


class BaseMetadataHook(HookModel, MetadataHookInterface, ABC):
    dependencies: Dependencies = Field(default_factory=dict)
    optional_dependencies: dict[str, Dependencies] = Field(default_factory=dict)
    path: GradlePath

    @abstractmethod
    def get_format_value(self, key: str) -> Any | None: ...

    @abstractmethod
    def parse_gradle_dependency(self, dependency: GradleDependency) -> str: ...

    def update(self, metadata: dict[str, Any]) -> None:
        """Implements MetadataHookInterface."""
        self.set_dynamic(
            metadata,
            "dependencies",
            self.parse_dependencies(self.dependencies),
        )

        self.set_dynamic(
            metadata,
            "optional-dependencies",
            {
                key: self.parse_dependencies(value)
                for key, value in self.optional_dependencies.items()
            },
        )

    @listify
    def parse_dependencies(self, dependencies: Dependencies) -> Iterator[str]:
        for dependency in dependencies:
            match dependency:
                case str():
                    result = ""
                    index = 0
                    for match in PLACEHOLDER_REGEX.finditer(dependency):
                        # if we fail to find a placeholder, just leave it in the string
                        # this is especially important for eg. Hatch's `{root:uri}`
                        if value := self.get_format_value(match.group(1)):
                            # add everything before the placeholder, then the value
                            result += dependency[index : match.start()] + value
                            # move the index to the character after the placeholder
                            index = match.end()
                    # add anything after the final placeholder
                    result += dependency[index:]
                    yield result
                case GradleDependency():
                    yield self.parse_gradle_dependency(dependency)

    def set_dynamic(self, metadata: dict[str, Any], key: str, value: Any):
        if key in metadata:
            raise ValueError(
                f"`{key}` may not be listed in the `project` table when using hatch-gradle-version to populate dependencies. Please use `tool.hatch.metadata.hooks.{self.PLUGIN_NAME}.{key}` instead."
            )
        if key not in metadata.get("dynamic", []):
            raise ValueError(
                f"`{key}` must be listed in `project.dynamic` when using hatch-gradle-version to populate dependencies."
            )
        metadata[key] = value



import re
from abc import ABC, abstractmethod
from typing import Any, TypedDict

from casefy import casefy
from hatchling.version.core import DEFAULT_PATTERN
from hatchling.version.source.plugin.interface import VersionSourceInterface
from pydantic import AliasChoices, Field

from hatch_gradle_version.common.codegen import write_code
from hatch_gradle_version.common.gradle import GradleVersion
from hatch_gradle_version.common.model import HookModel, KebabModel, ProjectPath

PY_VERSION_REGEX = re.compile(
    r'(?i)^(PY_VERSION) *= *([\'"])v?(?P<version>.+?)\2',
    flags=re.MULTILINE,
)

DEFAULT_REGEX = re.compile(DEFAULT_PATTERN, flags=re.MULTILINE)


class VersionData(TypedDict):
    version: str
    gradle_version: str
    py_version: str
    extra_versions: dict[str, str]
    full_gradle_version: GradleVersion


class GradleVersionRegex(KebabModel):
    pattern: re.Pattern[str] = Field(
        validation_alias=AliasChoices("pattern", "pat", "regex"),
    )
    replacement: str = Field(
        r"\1",
        validation_alias=AliasChoices("replacement", "repl"),
    )


class BaseVersionSource(HookModel, VersionSourceInterface, ABC):
    source: str
    py_path: ProjectPath
    scheme: str | None = None
    gradle_version_regex: re.Pattern[str] | GradleVersionRegex | None = None
    """If this is a standalone regex, it should contain exactly one match group,
    representing the part of the raw Gradle version to keep."""

    @abstractmethod
    def get_gradle_version(self) -> GradleVersion: ...

    def fmt_raw_gradle_version(self, raw: str, extra: dict[str, str]) -> str:
        match self.gradle_version_regex:
            case re.Pattern() as pattern:
                parsed = pattern.sub(r"\1", raw)
            case GradleVersionRegex(pattern=pattern, replacement=replacement):
                parsed = pattern.sub(replacement, raw)
            case None:
                parsed = raw
        return parsed.format_map(extra)

    def get_version_data(self):
        gradle_version = self.get_gradle_version()
        py_version = self.get_py_version()

        version = gradle_version.full_version(py_version)

        version_data = VersionData(
            version=version,
            gradle_version=str(gradle_version),
            py_version=py_version,
            extra_versions=gradle_version.extra_versions,
            full_gradle_version=gradle_version,
        )

        # write here because otherwise the other version constants get outdated
        self.set_version(version, version_data)
        return dict(version_data)

    def get_py_version(self) -> str:
        # load py_version from a file
        contents = self.py_path.read_text("utf-8")

        match = PY_VERSION_REGEX.search(contents) or DEFAULT_REGEX.search(contents)
        if not match:
            raise ValueError(
                f"Unable to parse the version from the file: {self.py_path}"
            )

        return match["version"]

    def set_version(self, version: str, version_data: VersionData | dict[str, Any]):
        # __version__.py
        write_code(
            self.py_path,
            f"""\
            # This file is auto-generated by hatch-gradle-version.
            # Only the value of PY_VERSION is editable. Do not edit other values.

            PY_VERSION = "{version_data["py_version"]}"
            """,
        )

        # __gradle_version__.py
        write_code(
            self.py_path.with_name("__gradle_version__.py"),
            f"""\
            # This file is auto-generated by hatch-gradle-version. Do not edit.

            GRADLE_VERSION = "{version_data["gradle_version"]}"
            FULL_VERSION = "{version}"
            """,
            *(
                f"{format_key(key, 'GRADLE_VERSION', 'FULL_VERSION')}={value!r}"
                for key, value in sorted(version_data["extra_versions"].items())
            ),
        )


def format_key(key: str, *reserved: str):
    key = casefy.constcase(key)
    if key in reserved:
        return f"{key}_"
    return key



import copy
from pathlib import Path
from textwrap import dedent

import pytest
from pytest import MonkeyPatch

from .version_catalog import VersionCatalogMetadataHook


def test_gradle_properties_deps(tmp_path: Path, monkeypatch: MonkeyPatch):
    # arrange
    monkeypatch.setenv("HATCH_GRADLE_DIR", "gradle_dir")

    version_catalog = tmp_path / "gradle_dir" / "gradle" / "my-libs.versions.toml"
    version_catalog.parent.mkdir(parents=True)
    version_catalog.write_text(
        dedent(
            """\
            [versions]
            K = "1.2.3"
            other = { strictly="1.0" }
            """
        )
    )

    hook = VersionCatalogMetadataHook(
        tmp_path.as_posix(),
        {
            "path": "gradle/my-libs.versions.toml",
            "dependencies": [
                {
                    "package": "P",
                    "op": "~=",
                    "key": "K",
                    "py-version": "4.5",
                }
            ],
        },
    )

    orig_metadata = {
        "dynamic": [
            "dependencies",
            "optional-dependencies",
        ],
    }

    # act
    metadata = copy.deepcopy(orig_metadata)
    hook.update(metadata)

    # assert
    assert metadata == orig_metadata | {
        "dependencies": ["P~=1.2.3.4.5"],
        "optional-dependencies": {},
    }


def test_non_str_version_fails(tmp_path: Path, monkeypatch: MonkeyPatch):
    # arrange
    monkeypatch.setenv("HATCH_GRADLE_DIR", "gradle_dir")

    version_catalog = tmp_path / "gradle_dir" / "gradle" / "my-libs.versions.toml"
    version_catalog.parent.mkdir(parents=True)
    version_catalog.write_text(
        dedent(
            """\
            [versions]
            K = { strictly="1.0" }
            """
        )
    )

    hook = VersionCatalogMetadataHook(
        tmp_path.as_posix(),
        {
            "path": "gradle/my-libs.versions.toml",
            "dependencies": [
                {
                    "package": "P",
                    "op": "~=",
                    "key": "K",
                    "py-version": "4.5",
                }
            ],
        },
    )

    metadata = {
        "dynamic": [
            "dependencies",
            "optional-dependencies",
        ],
    }

    # act
    with pytest.raises(ValueError):
        hook.update(metadata)
